import 'package:after_layout/after_layout.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:osam/osam.dart';

import 'navigation_presenter_interface.dart';

//todo: add cupertino routing

typedef Widget ScreenBuilder<T>(ValueKey<T> key, BuildContext ctx);

class OsamNavigator<T> extends StatefulWidget {
  final Map<T, ScreenBuilder<T>> screens;
  final INavigation<T> presenter;
  final bool rootNavigator;
  final T initialRoute;

  const OsamNavigator(
      {Key key, this.screens, this.presenter, this.rootNavigator = false, this.initialRoute})
      : super(key: key);

  @override
  _OsamNavigatorState createState() => _OsamNavigatorState<T>();
}

class _OsamNavigatorState<T> extends State<OsamNavigator<T>> with AfterLayoutMixin {
  final List<T> _routes = [];

  Map<T, ScreenBuilder> get screens => widget.screens;

  INavigation<T> get presenter => widget.presenter;

  @override
  Widget build(BuildContext context) =>
      widget.screens[widget.initialRoute](ValueKey(widget.initialRoute), context);

  void _silentPop(dynamic route) {
    if (presenter.routeStream.initialData.length > 1) {
      _routes.removeLast();
      presenter.pop();
    }
  }

  @override
  void afterFirstLayout(BuildContext context) {
    final navigator = Navigator.of(context, rootNavigator: widget.rootNavigator);
    final initialRoutes = presenter.routeStream.initialData;
    _routes.addAll(initialRoutes);
    if (initialRoutes.length > 1) {
      final firstRoute = initialRoutes.first;
      navigator.pushReplacement(CupertinoPageRoute(
          builder: ((ctx) => screens[firstRoute](ValueKey(firstRoute), ctx)),
          settings: RouteSettings(isInitialRoute: true)));
      final others = initialRoutes.sublist(1, initialRoutes.length);
      others.forEach((route) {
        navigator.push(CupertinoPageRoute(builder: ((ctx) => screens[route](ValueKey(route), ctx)))
          ..popped.then(_silentPop));
      });
    }
    presenter.routeStream.stream.listen((routes) {
      final routesDifferentNumber = routes.length - _routes.length;
      if (routesDifferentNumber == 0 && routes.last != _routes.last) {
        navigator.pushReplacement(CupertinoPageRoute(builder: ((ctx) {
          return screens[routes.last](ValueKey(routes.last), ctx);
        }))
          ..popped.then(_silentPop));
      } else if (routesDifferentNumber > 0) {
        final pushRoutes = routes.sublist(_routes.length, routes.length);
        pushRoutes.forEach((route) {
          navigator.push(
              CupertinoPageRoute(builder: ((ctx) => screens[route](ValueKey(route), ctx)))
                ..popped.then(_silentPop));
        });
      } else if (routesDifferentNumber < 0) {
        _routes.first != routes.first
            ? navigator.pushAndRemoveUntil(CupertinoPageRoute(builder: ((ctx) {
                return screens[routes.first](ValueKey(routes.first), ctx);
              })), (route) => false)
            : _routes.sublist(routes.length, _routes.length).forEach((route) async {
                navigator.pop();
              });
      }
      _routes.clear();
      _routes.addAll(routes);
    });
  }
}

class OsamTabNavigator<T> extends StatelessWidget {
  final Map<T, ScreenBuilder<T>> screens;
  final Map<T, Widget> tabs;
  final ITabNavigation<T> presenter;
  final BoxDecoration tabDecoration;

  OsamTabNavigator({Key key, this.screens, this.presenter, this.tabDecoration, this.tabs})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    return OsamBuilder<T>(
      stream: presenter.tabStream,
      builder: (ctx, snapshot) {
        return Stack(
          children: <Widget>[
            IndexedStack(
              index: screens.keys.toList().indexOf(snapshot.data),
              children: screens.keys.map((tab) {
                return _OsamNavigatorPage(
                  key: ValueKey(tab.toString() + '_onp'),
                  navKey: GlobalKey<NavigatorState>(),
                  child: screens[tab](ValueKey(tab), context),
                );
              }).toList(),
            ),
            Positioned(
              bottom: 0,
              child: SafeArea(
                bottom: true,
                child: Container(
                  decoration: tabDecoration,
                  padding: const EdgeInsets.symmetric(vertical: 5),
                  constraints: BoxConstraints(
                    maxWidth: MediaQuery.of(context).size.width,
                    minHeight: 48,
                  ),
                  child: Row(
                    children: screens.keys
                        .map((tab) => Expanded(
                              child: GestureDetector(
                                child: Container(
                                  child: tabs[tab],
                                  alignment: Alignment.center,
                                  color: Colors.transparent,
                                ),
                                onTap: () {
                                  if (snapshot.data != tab) {
                                    presenter.changeTab(tab);
                                  } else {
                                    // todo: pop to initial in current navigator
                                  }
                                },
                              ),
                            ))
                        .toList(),
                  ),
                ),
              ),
            )
          ],
        );
      },
    );
  }
}

//onTap: () {
//if (currentTab != enumTab) {
//routeTo(enumTab);
//} else {
//final routes = currentPresenter.initialRoutes;
//if (routes.length > 1) {
//routes.sublist(1, routes.length).forEach((route) {
//currentPresenter.pop();

class _OsamNavigatorPage extends StatelessWidget {
  final GlobalKey<NavigatorState> navKey;
  final Widget child;

  const _OsamNavigatorPage({Key key, this.navKey, this.child}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Navigator(
      key: navKey,
      onGenerateRoute: (RouteSettings settings) => MaterialPageRoute(
        settings: RouteSettings(isInitialRoute: true),
        builder: (BuildContext context) => child,
      ),
    );
  }
}
