import 'package:flutter/material.dart';
import 'package:osam/domain/state/base_state.dart';
import 'package:osam/domain/store/store.dart';
import 'package:osam/presentation/presenter.dart';
import 'package:provider/provider.dart';

typedef void LifeCycleHandler(AppLifecycleState state);

class StoreProvider<ST extends BaseState<ST>> extends StatelessWidget {
  final Store<ST> store;
  final Widget child;
  final LifeCycleHandler lifeCycleHandler;

  const StoreProvider({Key key, @required this.store, @required this.child, this.lifeCycleHandler})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    return _InternalStoreProvider(
        store: store, child: store.isPersistEnabled ? _LifecycleWrapper<ST>(child: child) : child);
  }
}

class _InternalStoreProvider extends InheritedWidget {
  const _InternalStoreProvider({
    Key key,
    @required this.child,
    @required this.store,
  }) : super(key: key, child: child);

  final Store store;
  final Widget child;

  static _InternalStoreProvider of(BuildContext context) =>
      context.dependOnInheritedWidgetOfExactType(aspect: _InternalStoreProvider);

  @override
  bool updateShouldNotify(_InternalStoreProvider oldWidget) => this.store != oldWidget.store;
}

class _LifecycleWrapper<ST extends BaseState<ST>> extends StatefulWidget {
  final Widget child;
  final LifeCycleHandler lifeCycleHandler;

  const _LifecycleWrapper({Key key, this.child, this.lifeCycleHandler}) : super(key: key);

  @override
  _LifecycleWrapperState createState() => _LifecycleWrapperState<ST>();
}

class _LifecycleWrapperState<ST extends BaseState<ST>> extends State<_LifecycleWrapper>
    with WidgetsBindingObserver {
  @override
  void initState() {
    WidgetsBinding.instance.addObserver(this);
    super.initState();
  }

  @override
  Widget build(BuildContext context) => widget.child;

  @override
  void didChangeAppLifecycleState(AppLifecycleState state) {
    if (state == AppLifecycleState.paused ||
        state == AppLifecycleState.inactive ||
        state == AppLifecycleState.detached) _InternalStoreProvider.of(context).store.storeState();
    widget.lifeCycleHandler(state);
  }
}

class PresenterProvider<P extends Presenter<BaseState>> extends StatelessWidget {
  final P presenter;
  final Widget child;

  const PresenterProvider({Key key, @required this.presenter, @required this.child})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    final store = _InternalStoreProvider.of(context).store;
    assert(store != null);
    if (store == null)
      debugPrint('insure you provided store by StoreProvider in the start of your widget tree');
    return Provider(
      key: ValueKey('PresenterProvider: %${key.toString()}'),
      create: (_) => presenter
        // ignore: invalid_use_of_protected_member
        ..store = store,
      child: child,
      dispose: (ctx, presenter) => presenter.dispose(),
    );
  }

  static P of<P extends Presenter<BaseState>>(BuildContext context) =>
      Provider.of<P>(context, listen: false);
}

class PresenterMultiProvider<P extends Presenter<BaseState>> extends StatelessWidget {
  final List<P> presenters;
  final Widget child;

  const PresenterMultiProvider({Key key, @required this.presenters, @required this.child})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    final store = _InternalStoreProvider.of(context).store;
    assert(store != null);
    if (store == null)
      debugPrint('insure you provided store by StoreProvider in the start of your widget tree');
    return MultiProvider(
      providers: presenters
          .map((presenter) => Provider(
                create: (_) => presenter,
                dispose: (ctx, presenter) => presenter.dispose(),
              ))
          .toList(),
      key: ValueKey('PresenterMultiProvider: %${key.toString()}'),
      child: child,
    );
  }

  static P of<P extends Presenter<BaseState>>(BuildContext context) =>
      PresenterMultiProvider.of<P>(context);
}
